@@ -182,7 +182,7 @@ struct cbinfo {
 
 /* s3backer_store functions */
 static int block_cache_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *block_sizep);
-static int block_cache_set_mounted(struct s3backer_store *s3b, int *old_valuep, int new_value);
+static int block_cache_set_mount_token(struct s3backer_store *s3b, int32_t *old_valuep, int32_t new_value);
 static int block_cache_read_block(struct s3backer_store *s3b, s3b_block_t block_num, void *dest,
   u_char *actual_md5, const u_char *expect_md5, int strict);
 static int block_cache_write_block(struct s3backer_store *s3b, s3b_block_t block_num, const void *src, u_char *md5,
@@ -244,7 +244,7 @@ block_cache_create(struct block_cache_conf *config, struct s3backer_store *inner
         goto fail0;
     }
     s3b->meta_data = block_cache_meta_data;
-    s3b->set_mounted = block_cache_set_mounted;
+    s3b->set_mount_token = block_cache_set_mount_token;
     s3b->read_block = block_cache_read_block;
     s3b->write_block = block_cache_write_block;
     s3b->read_block_part = block_cache_read_block_part;
@@ -290,9 +290,11 @@ block_cache_create(struct block_cache_conf *config, struct s3backer_store *inner
     /* Initialize on-disk cache and read in directory */
     if (config->cache_file != NULL) {
         if ((r = s3b_dcache_open(&priv->dcache, config->log, config->cache_file, config->block_size,
-          config->cache_size, block_cache_dcache_load, priv)) != 0)
+          config->cache_size, block_cache_dcache_load, priv, config->perform_flush)) != 0)
             goto fail9;
-        priv->stats.initial_size = priv->num_cleans;
+        if (config->perform_flush && priv->num_dirties > 0)
+            (*config->log)(LOG_INFO, "%u dirty blocks in cache file `%s' will be recovered", priv->num_dirties, config->cache_file);
+        priv->stats.initial_size = priv->num_cleans + priv->num_dirties;
     }
 
     /* Grab lock */
@@ -352,6 +354,7 @@ fail0:
 static int
 block_cache_dcache_load(void *arg, s3b_block_t dslot, s3b_block_t block_num, const u_char *md5)
 {
+    const u_int dirty = md5 == NULL;
     struct block_cache_private *const priv = arg;
     struct block_cache_conf *const config = priv->config;
     struct cache_entry *entry;
@@ -359,6 +362,7 @@ block_cache_dcache_load(void *arg, s3b_block_t dslot, s3b_block_t block_num, con
 
     /* Sanity check */
     assert(config->cache_file != NULL);
+    assert(!dirty || config->perform_flush);            /* we should never see dirty blocks unless we asked for them */
 
     /* Sanity check a block is not listed twice */
     if ((entry = s3b_hash_get(priv->hashtable, block_num)) != NULL) {
@@ -367,7 +371,7 @@ block_cache_dcache_load(void *arg, s3b_block_t dslot, s3b_block_t block_num, con
         return EINVAL;
     }
 
-    /* Create a new cache entry in state CLEAN[2] */
+    /* Create a new cache entry */
     assert(config->cache_file != NULL);
     if ((entry = calloc(1, sizeof(*entry) + (!config->no_verify ? MD5_DIGEST_LENGTH : 0))) == NULL) {
         r = errno;
@@ -376,15 +380,24 @@ block_cache_dcache_load(void *arg, s3b_block_t dslot, s3b_block_t block_num, con
         return r;
     }
     entry->block_num = block_num;
-    entry->verify = !config->no_verify;
     entry->timeout = block_cache_get_time(priv) + priv->clean_timeout;
-    if (entry->verify)
-        memcpy(&entry->md5, md5, MD5_DIGEST_LENGTH);
     entry->u.dslot = dslot;
-    TAILQ_INSERT_TAIL(&priv->cleans, entry, link);
-    priv->num_cleans++;
+
+    /* Mark as clean or dirty accordingly */
+    if (dirty) {
+        entry->dirty = 1;
+        TAILQ_INSERT_TAIL(&priv->dirties, entry, link);
+        priv->num_dirties++;
+        assert(ENTRY_GET_STATE(entry) == DIRTY);
+    } else {
+        entry->verify = !config->no_verify;
+        if (entry->verify)
+            memcpy(&entry->md5, md5, MD5_DIGEST_LENGTH);
+        TAILQ_INSERT_TAIL(&priv->cleans, entry, link);
+        priv->num_cleans++;
+        assert(ENTRY_GET_STATE(entry) == (config->no_verify ? CLEAN : CLEAN2));
+    }
     s3b_hash_put_new(priv->hashtable, entry);
-    assert(ENTRY_GET_STATE(entry) == (config->no_verify ? CLEAN : CLEAN2));
     return 0;
 }
 
@@ -397,11 +410,21 @@ block_cache_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *bloc
 }
 
 static int
-block_cache_set_mounted(struct s3backer_store *s3b, int *old_valuep, int new_value)
+block_cache_set_mount_token(struct s3backer_store *s3b, int32_t *old_valuep, int32_t new_value)
 {
     struct block_cache_private *const priv = s3b->data;
+    int r;
 
-    return (*priv->inner->set_mounted)(priv->inner, old_valuep, new_value);
+    /* Set flag in lower layer */
+    if ((r = (*priv->inner->set_mount_token)(priv->inner, old_valuep, new_value)) != 0)
+        return r;
+
+    /* Update the disk cache file as well, if the value was changed */
+    if (priv->dcache != NULL && new_value >= 0)
+        r = s3b_dcache_set_mount_token(priv->dcache, NULL, new_value);
+
+    /* Done */
+    return 0;
 }
 
 static int
@@ -784,10 +807,10 @@ again:
                 goto again;
             }
 
-            /* Invalidate disk cache entry */
+            /* Record dirty disk cache entry */
             if (config->cache_file != NULL) {
-                if ((r = s3b_dcache_erase_block(priv->dcache, entry->u.dslot)) != 0)
-                    (*config->log)(LOG_ERR, "can't erase cached block! %s", strerror(r));
+                if ((r = s3b_dcache_record_block(priv->dcache, entry->u.dslot, entry->block_num, NULL)) != 0)
+                    (*config->log)(LOG_ERR, "can't dirty cached block %u! %s", block_num,  strerror(r));
             }
 
             /* Change from CLEAN to DIRTY */
@@ -855,6 +878,12 @@ again:
     priv->num_dirties++;
     assert(ENTRY_GET_STATE(entry) == DIRTY);
 
+    /* Record dirty disk cache entry */
+    if (config->cache_file != NULL) {
+        if ((r = s3b_dcache_record_block(priv->dcache, entry->u.dslot, entry->block_num, NULL)) != 0)
+            (*config->log)(LOG_ERR, "can't dirty cached block %u! %s", block_num,  strerror(r));
+    }
+
     /* Wake up a worker thread to go write it */
     pthread_cond_signal(&priv->worker_work);
 
