@@ -41,6 +41,7 @@
 #include "http_io.h"
 #include "test_io.h"
 #include "s3b_config.h"
+#include "dcache.h"
 
 /****************************************************************************
  *                          DEFINITIONS                                     *
@@ -82,9 +83,6 @@
 #define S3BACKER_DEFAULT_COMPRESSION                Z_NO_COMPRESSION
 #define S3BACKER_DEFAULT_ENCRYPTION                 "AES-128-CBC"
 
-#define S3BACKER_SSE_HEADER                         "x-amz-server-side-encryption"
-#define S3BACKER_DEFAULT_SSE_VALUE                  "AES256"
-
 /* MacFUSE setting for kernel daemon timeout */
 #ifdef __APPLE__
 #ifndef FUSE_MAX_DAEMON_TIMEOUT
@@ -152,6 +150,7 @@ static struct s3b_config config = {
         .baseURL=               NULL,
         .region=                NULL,
         .bucket=                NULL,
+        .sse=                   NULL,
         .prefix=                S3BACKER_DEFAULT_PREFIX,
         .accessType=            S3BACKER_DEFAULT_ACCESS_TYPE,
         .authVersion=           S3BACKER_DEFAULT_AUTH_VERSION,
@@ -160,7 +159,6 @@ static struct s3b_config config = {
         .timeout=               S3BACKER_DEFAULT_TIMEOUT,
         .initial_retry_pause=   S3BACKER_DEFAULT_INITIAL_RETRY_PAUSE,
         .max_retry_pause=       S3BACKER_DEFAULT_MAX_RETRY_PAUSE,
-        .sse =                  NULL
     },
 
     /* "Eventual consistency" protection config */
@@ -243,8 +241,8 @@ static const struct fuse_opt option_list[] = {
         .offset=    offsetof(struct s3b_config, http_io.region),
     },
     {
-            .templ=     "--sse=%s",
-            .offset=    offsetof(struct s3b_config, http_io.sse),
+        .templ=     "--sse=%s",
+        .offset=    offsetof(struct s3b_config, http_io.sse),
     },
     {
         .templ=     "--blockCacheSize=%u",
@@ -271,6 +269,11 @@ static const struct fuse_opt option_list[] = {
         .templ=     "--blockCacheMaxDirty=%u",
         .offset=    offsetof(struct s3b_config, block_cache.max_dirty),
     },
+    {
+        .templ=     "--blockCacheRecoverDirtyBlocks",
+        .offset=    offsetof(struct s3b_config, block_cache.recover_dirty_blocks),
+        .value=     1
+    },
     {
         .templ=     "--readAhead=%u",
         .offset=    offsetof(struct s3b_config, block_cache.read_ahead),
@@ -596,7 +599,8 @@ struct s3backer_store *
 s3backer_create_store(struct s3b_config *conf)
 {
     struct s3backer_store *store;
-    int mounted;
+    int32_t old_mount_token;
+    int32_t new_mount_token;
     int r;
 
     /* Sanity check */
@@ -630,19 +634,28 @@ s3backer_create_store(struct s3b_config *conf)
         store = block_cache_store;
     }
 
-    /* Set mounted flag and check previous value one last time */
-    r = (*store->set_mounted)(store, &mounted, conf->fuse_ops.read_only ? -1 : 1);
-    if (r != 0) {
-        (*conf->log)(LOG_ERR, "error reading mounted flag on %s: %s", conf->description, strerror(r));
+    /* Set mount token and check previous value one last time */
+    new_mount_token = -1;
+    if (!conf->fuse_ops.read_only) {
+        srandom((long)time(NULL) ^ (long)&old_mount_token);
+        do
+            new_mount_token = random();
+        while (new_mount_token <= 0);
+    }
+    if ((r = (*store->set_mount_token)(store, &old_mount_token, new_mount_token)) != 0) {
+        (*conf->log)(LOG_ERR, "error reading mount token on %s: %s", conf->description, strerror(r));
         goto fail;
     }
-    if (mounted) {
-        if (!conf->force) {
-            (*conf->log)(LOG_ERR, "%s appears to be mounted by another s3backer process", config.description);
+    if (old_mount_token != 0) {
+        if (!conf->force && !conf->block_cache.perform_flush) {
+            (*conf->log)(LOG_ERR, "%s appears to be mounted by another s3backer process (using mount token 0x%08x)",
+              config.description, (int)old_mount_token);
             r = EBUSY;
             goto fail;
         }
     }
+    if (new_mount_token != -1)
+        (*conf->log)(LOG_INFO, "established new mount token 0x%08x", (int)new_mount_token);
 
     /* Done */
     return store;
@@ -1016,11 +1029,10 @@ validate_config(void)
         return -1;
     }
 
-    /* Check storage class */
-    if (config.http_io.sse != NULL
-        && strcmp(config.http_io.sse, S3BACKER_DEFAULT_SSE_VALUE) != 0) {
-        warnx("invalid sse type `%s'", config.http_io.sse);
-        config.http_io.sse = S3BACKER_DEFAULT_SSE_VALUE;
+    /* Check server side encryption type */
+    if (config.http_io.sse != NULL && strcmp(config.http_io.sse, REQUIRED_SSE_VALUE) != 0) {
+        warnx("invalid sse type `%s' (only `%s' is supported)", config.http_io.sse, REQUIRED_SSE_VALUE);
+        return -1;
     }
 
     /* Set default or custom region */
@@ -1262,6 +1274,10 @@ validate_config(void)
             return -1;
         }
     }
+    if (config.block_cache.cache_file == NULL && config.block_cache.recover_dirty_blocks) {
+        warnx("`--blockCacheRecoverDirtyBlocks' requires specifying `--blockCacheFile'");
+        return -1;
+    }
 
     /* Check mount point */
     if (config.erase || config.reset) {
@@ -1368,31 +1384,6 @@ validate_config(void)
         break;
     }
 
-    /* Check whether already mounted */
-    if (!config.test && !config.erase && !config.reset) {
-        int mounted;
-
-        config.http_io.debug = config.debug;
-        config.http_io.quiet = config.quiet;
-        config.http_io.log = config.log;
-        if ((s3b = http_io_create(&config.http_io)) == NULL)
-            err(1, "http_io_create");
-        r = (*s3b->set_mounted)(s3b, &mounted, -1);
-        (*s3b->destroy)(s3b);
-        if (r != 0) {
-            errno = r;
-            err(1, "error reading mounted flag");
-        }
-        if (mounted) {
-            if (!config.force)
-                errx(1, "error: %s appears to be already mounted", config.description);
-            if (!config.quiet) {
-                warnx("warning: filesystem appears already mounted but you said `--force'\n"
-                  " so I'll proceed anyway even though your data may get corrupted.\n");
-            }
-        }
-    }
-
     /* Check computed block and file sizes */
     if (config.block_size != (1 << (ffs(config.block_size) - 1))) {
         warnx("block size must be a power of 2");
@@ -1481,6 +1472,89 @@ validate_config(void)
     config.fuse_ops.num_blocks = config.num_blocks;
     config.fuse_ops.log = config.log;
 
+    /* Check whether already mounted, and if so, compare mount token against on-disk cache (if any) */
+    if (!config.test && !config.erase && !config.reset) {
+        int32_t mount_token;
+        int conflict;
+
+        /* Read s3 mount token */
+        config.http_io.debug = config.debug;
+        config.http_io.quiet = config.quiet;
+        config.http_io.log = config.log;
+        if ((s3b = http_io_create(&config.http_io)) == NULL)
+            err(1, "http_io_create");
+        r = (*s3b->set_mount_token)(s3b, &mount_token, -1);
+        (*s3b->destroy)(s3b);
+        if (r != 0) {
+            errno = r;
+            err(1, "error reading mount token");
+        }
+        conflict = mount_token != 0;
+
+        /*
+         * The disk cache also has a mount token, so we need to do some extra checking.
+         * Either token can be 0 (i.e., not present -> not mounted) or != 0 (mounted).
+         *
+         * If neither token is present, proceed with mount. Note: there should not be
+         * any dirty blocks in the disk cache in this case, because this represents a
+         * clean unmount situation.
+         *
+         * If the cache has a token, but S3 has none, that means someone must have used
+         * `--reset-mounted-flag' to clear it from S3 since the last time the disk cache was
+         * used. In that case, `--force' is required to continue using the disk cache,
+         * or `--reset-mounted-flag' must be used to clear the disk cache flag as well.
+         *
+         * If --blockCacheRecoverDirtyBlocks is specified and the tokens match, we
+         * have the corresponding cache file for the last mount. Proceed with mount and,
+         * if configured, enable cache writeback of dirty blocks.
+         */
+        if (config.block_cache.cache_file != NULL) {
+            int32_t cache_mount_token = -1;
+            struct stat cache_file_stat;
+            struct s3b_dcache *dcache;
+
+            /* Open disk cache file, if any, and read the mount token therein, if any */
+            if (stat(config.block_cache.cache_file, &cache_file_stat) == -1) {
+                if (errno != ENOENT)
+                    err(1, "can't open cache file `%s'", config.block_cache.cache_file);
+            } else {
+                if ((r = s3b_dcache_open(&dcache, config.log, config.block_cache.cache_file,
+                  config.block_cache.block_size, config.block_cache.cache_size, NULL, NULL, 0)) != 0)
+                    errx(1, "error opening cache file `%s': %s", config.block_cache.cache_file, strerror(r));
+                if (s3b_dcache_has_mount_token(dcache) && (r = s3b_dcache_set_mount_token(dcache, &cache_mount_token, -1)) != 0)
+                    errx(1, "error reading mount token from `%s': %s", config.block_cache.cache_file, strerror(r));
+                s3b_dcache_close(dcache);
+            }
+
+            /* If cache file is older format, then cache_mount_token will be -1, otherwise >= 0 */
+            if (cache_mount_token > 0) {
+
+                /* If tokens do not agree, bail out, otherwise enable write-back of dirty blocks if tokens are non-zero */
+                if (cache_mount_token != mount_token) {
+                    warnx("cache file `%s' mount token mismatch (disk:0x%08x != s3:0x%08x)",
+                      config.block_cache.cache_file, cache_mount_token, mount_token);
+                } else if (config.block_cache.recover_dirty_blocks) {
+                    if (!config.quiet)
+                        warnx("recovering from unclean shutdown: dirty blocks in cache file will be written back to S3");
+                    config.block_cache.perform_flush = 1;
+                    conflict = 0;
+                }
+            }
+        }
+
+        /* If there is a conflicting mount, additional `--force' is required */
+        if (conflict) {
+            if (!config.force) {
+                warnx("%s appears to be already mounted (using mount token 0x%08x)", config.description, (int)mount_token);
+                errx(1, "reset mount token with `--reset-mounted-flag', or use `--force' to override");
+            }
+            if (!config.quiet) {
+                warnx("warning: filesystem appears already mounted but you said `--force'\n"
+                  " so I'll proceed anyway even though your data may get corrupted.\n");
+            }
+        }
+    }
+
     /* If `--listBlocks' was given, build non-empty block bitmap */
     if (config.erase || config.reset)
         config.list_blocks = 0;
@@ -1584,6 +1658,7 @@ dump_config(void)
       config.max_speed_str[HTTP_DOWNLOAD] != NULL ? config.max_speed_str[HTTP_DOWNLOAD] : "-",
       config.http_io.max_speed[HTTP_DOWNLOAD]);
     (*config.log)(LOG_DEBUG, "%24s: %us", "timeout", config.http_io.timeout);
+    (*config.log)(LOG_DEBUG, "%24s: \"%s\"", "sse", config.http_io.sse);
     (*config.log)(LOG_DEBUG, "%24s: %ums", "initial_retry_pause", config.http_io.initial_retry_pause);
     (*config.log)(LOG_DEBUG, "%24s: %ums", "max_retry_pause", config.http_io.max_retry_pause);
     (*config.log)(LOG_DEBUG, "%24s: %ums", "min_write_delay", config.ec_protect.min_write_delay);
@@ -1595,6 +1670,7 @@ dump_config(void)
     (*config.log)(LOG_DEBUG, "%24s: %ums", "block_cache_write_delay", config.block_cache.write_delay);
     (*config.log)(LOG_DEBUG, "%24s: %u blocks", "block_cache_max_dirty", config.block_cache.max_dirty);
     (*config.log)(LOG_DEBUG, "%24s: %s", "block_cache_sync", config.block_cache.synchronous ? "true" : "false");
+    (*config.log)(LOG_DEBUG, "%24s: %s", "recover_dirty_blocks", config.block_cache.recover_dirty_blocks ? "true" : "false");
     (*config.log)(LOG_DEBUG, "%24s: %u blocks", "read_ahead", config.block_cache.read_ahead);
     (*config.log)(LOG_DEBUG, "%24s: %u blocks", "read_ahead_trigger", config.block_cache.read_ahead_trigger);
     (*config.log)(LOG_DEBUG, "%24s: \"%s\"", "block_cache_cache_file",
@@ -1696,6 +1772,7 @@ usage(void)
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheNoVerify", "Disable verification of data loaded from cache file");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheSize=NUM", "Block cache size (in number of blocks)");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheSync", "Block cache performs all writes synchronously");
+    fprintf(stderr, "\t--%-27s %s\n", "blockCacheRecoverDirtyBlocks", "Recover dirty cache file blocks on startup");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheThreads=NUM", "Block cache write-back thread pool size");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheTimeout=MILLIS", "Block cache entry timeout (zero = infinite)");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheWriteDelay=MILLIS", "Block cache maximum write-back delay");
@@ -1711,7 +1788,7 @@ usage(void)
     fprintf(stderr, "\t--%-27s %s\n", "filename=NAME", "Name of backed file in filesystem");
     fprintf(stderr, "\t--%-27s %s\n", "force", "Ignore different auto-detected block and file sizes");
     fprintf(stderr, "\t--%-27s %s\n", "help", "Show this information and exit");
-    fprintf(stderr, "\t--%-27s %s\n", "initialRetryPause=MILLIS", "Inital retry pause after stale data or server error");
+    fprintf(stderr, "\t--%-27s %s\n", "initialRetryPause=MILLIS", "Initial retry pause after stale data or server error");
     fprintf(stderr, "\t--%-27s %s\n", "insecure", "Don't verify SSL server identity");
     fprintf(stderr, "\t--%-27s %s\n", "keyLength", "Override generated cipher key length");
     fprintf(stderr, "\t--%-27s %s\n", "listBlocks", "Auto-detect non-empty blocks at startup");
@@ -1733,6 +1810,7 @@ usage(void)
     fprintf(stderr, "\t--%-27s %s\n", "reset-mounted-flag", "Reset `already mounted' flag in the filesystem");
     fprintf(stderr, "\t--%-27s %s\n", "rrs", "Target written blocks for Reduced Redundancy Storage (deprecated)");
     fprintf(stderr, "\t--%-27s %s\n", "size=SIZE", "File size (with optional suffix 'K', 'M', 'G', etc.)");
+    fprintf(stderr, "\t--%-27s %s\n", "sse=" REQUIRED_SSE_VALUE, "Specify server side encryption");
     fprintf(stderr, "\t--%-27s %s\n", "ssl", "Enable SSL");
     fprintf(stderr, "\t--%-27s %s\n", "statsFilename=NAME", "Name of statistics file in filesystem");
     fprintf(stderr, "\t--%-27s %s\n", "storageClass=TYPE", "Specify storage class for written blocks");
